#!/usr/bin/perl -w

# Written by Simon Huggins
# (C) Project Purple 1999
# 21/7/1999 - started
# 24/7/1999 - sorted out my sendmail config so I didn't have to be online to
# 	      test it and thus made some progress :-)
# v0.0.2
# 27/8/1999 - Used News::Article
 
use strict;
use English;
use Net::POP3;
use Net::SMTP;
use Carp;
use News::Article;

print "Filtermail v0.0.1 (C) Project Purple 1999 released under the GPL\n";

my ($pop, $msgsref, $nummsgs, $verbosity, $headerref, $do);
my ($smtp, $bodyref, $from, %config, @msgidcache, $idline);
my ($headers, $tmp);

# One action per regexp (when regexp matches return action)
my (@actions, @regexps, $num_actions, @removeactions);

$num_actions=0;

# Action	Meaning
# 
# 0		Leave on server, add to cache, get users idea.
# 1		Get.
# 2		Delete.
# 3		Leave, but already in cache.

sub getactions_from_file() {
	my $name=shift;

	if (open(INST, "<$name")) {
	while (<INST>) {
		if (/([0-9])\s+(.*)/) {
			push @actions, $1;
			push @regexps, $2;
			$num_actions++;
		}
	}
	close(INST);
	}
}

sub filter() {
	my $msgsize=shift;
	my @headers=@_;
	my ($line,$i);

	# Check headers against hash of headers to check.

	for $line (@headers) {
		for ($i=0; $i<$num_actions; $i++) {
			if ($line =~ /$regexps[$i]/) {
				push @removeactions, $actions[$i] . " " . $regexps[$i];
				return $actions[$i];
			}
		}
	}
	return 0;	# Default is to cache and await instructions
}

### Dienicely just um, dies nicely (given $pop)

sub dienicely() {
	my $message=shift;
	my $server=shift;
	chomp;
	if ($verbosity>2) {
		$_=$server->message();
		chomp;
		croak "$message (Server said \"$_\")\n";
	} else {
		croak "$message\n";
	}
}

### Output_header outputs the headers into CACHE with the linefeeds and
### headers themselves

sub output_header() {
	my $hdr=shift;
	my $tmp;

	if ($tmp=$headers->header($hdr)) {
		print CACHE $hdr . ": " . $tmp . "\n";
	}
}


### Do our own ARGV parsing cos diving into getopts looks like a pita
### just for one arg.

$verbosity=0;
for (@ARGV) {
	if (/-(v+)/) {
		$verbosity += length($1);
	} else {
		print "Invalid option $_ given on commandline!\n";
		croak "Usage: $0 [-v]\n";
	}
}

if ($verbosity>0) { print "Being verbose (" . $verbosity . ")\n"; }

### Should read a config file for now just stuff the hash

$config{'to'}="huggie\@localhost";
$config{'login'}="poptest";
$config{'password'}="secret";
$config{'server'}="localhost";

### Read the instructions for Message IDs

&getactions_from_file('pre_instructions'); 	# FIXME needs to be better
&getactions_from_file('instructions'); 		# algorithm really 
&getactions_from_file('post_instructions');	# should be clever/more
						# flexible.

### Read the cache and make up a cache of msgids - don't add anything to the
### cache twice

if (open(CACHE, '<cache')) {
	while (<CACHE>) {
		if (/Message-ID:/i) { push @msgidcache, $_; }
	}
	close(CACHE);
}

### Establish a POP3 connection

$pop=Net::POP3->new($config{'server'}, Timeout => 60);
unless ($pop) {
	croak("That went rather wrong then! Couldn't even initialise connection");
}

$nummsgs=$pop->login($config{'login'},$config{'password'});

unless ($pop->ok()) { 
	&dienicely("Couldn't authenticate $config{'login'}",$pop);
}
if ($nummsgs == 0) { croak "No mail for user poptest!\n"; }
if ($verbosity>1) { print "$nummsgs messages on server\n"; }

### Get the LIST of message number and length (octets)

$msgsref=$pop->list();
unless ($pop->ok()) {
	&dienicely("Couldn't get list of messages",$pop);
}
$|=1;
MSG: for (keys %$msgsref) {
	if ($verbosity>2) {
		#	Number	Size (in octets)
		print "$_\t" . $$msgsref{$_} . "\n";
	}
	$headerref=$pop->top($_, 0); 	# Send a TOP and return an arrayref to
					# the headers
	unless ($pop->ok()) {
		&dienicely("Couldn't get headers of message $_",$pop);
	}
	if ($verbosity>3) {
		print "Headers of message $_\n@$headerref";
	}
	
	$headers=News::Article->new($headerref);
	unless ($headers) {
		croak("Couldn't create News::Article");
	}
	# Call filter on the array of headers

	$do=&filter($_,@$headerref);

	# do the right thing by $do

	unless ($from=$headers->header('From')) {
		carp "Couldn't find From: header. Broken message?\nSetting
envelope from to $config{'to'}";
		$from=$config{'to'};
		# FIXME Add an error header too?
		}
	unless ($idline=$headers->header('Message-ID')) {
		carp "Couldn't find Message-ID: header. Broken message?";
		# FIXME Add an error header too?
		}

	if ($do==0) {
		my $num=$_;

		for (@msgidcache) {
			if (/^Message-ID: $idline$/) {
				print "Had already seen $num\n";
				next MSG;
			}
		}

		if ($verbosity>2) {
			print "Left $num and cached its ID\n";
		}
		# Leave and cache it
		open(CACHE, ">>cache");
		&output_header('In-Reply-To');
		&output_header('To');
		&output_header('From');
		&output_header('Message-ID');
		&output_header('References');
		&output_header('Organisation');
		&output_header('Organization');
		&output_header('Subject');
		&output_header('Sender');
		&output_header('X-Mailer');
		&output_header('Cc');
		print CACHE "---SIZE=$$msgsref{$num}\n";
		close CACHE;
		next;
	} elsif ($do==1) {
		if ($verbosity>2) {
			print "Retrieved $_\n";
		}
		# Get it
		$bodyref=$pop->get($_);
		unless ($pop->ok()) {
			&dienicely("Couldn't get message $_", $pop);
		}
		if (not defined $smtp) { 			# then open a connection
			$smtp=Net::SMTP->new('localhost', Timeout => 60,
							  Hello => 'localhost');
			unless ($smtp->ok()) {
				&dienicely("Couldn't log onto SMTP server", $smtp);
			}
			if ($verbosity>2) {
				print "Opened SMTP connection\n";
			}
		}

		$smtp->mail($from);
		unless ($smtp->ok()) {
			&dienicely("Couldn't send from", $smtp);
		}
		$smtp->to($config{'to'});
		unless ($smtp->ok()) {
			&dienicely("Couldn't send to", $smtp);
		}

		$smtp->data();
		unless ($smtp->ok()) {
			&dienicely("Couldn't initiate data connection", $smtp);
		}
		$smtp->datasend(@$bodyref);
		unless ($smtp->ok()) {
			&dienicely("Couldn't send data", $smtp);
		}
		$smtp->dataend();
		unless ($smtp->ok()) {
			&dienicely("Couldn't stop data connection", $smtp);
		}
		if ($verbosity>2) {
			print "Sent message $_\n";
		}
		#$pop->delete($_);
		#unless ($pop->ok()) {
		#	&dienicely("Couldn't delete message $_", $pop);
		#}
	} elsif ($do==2) {
		if ($verbosity>2) {
			print "Deleted $_\n";
		}
		# Delete it
		$pop->delete($_);
		unless ($pop->ok()) {
			&dienicely("Couldn't delete message $_", $pop);
		}
	} elsif ($do==3) {
		if ($verbosity>2) {
			print "Left $_ on server.\n";
		}
		# Leave it
		next;
	}
}

$pop->quit();
unless ($pop->ok()) {
	&dienicely("Couldn't logout from pop server", $pop);
}
if (defined $smtp) {
	$smtp->quit();
	unless ($smtp->ok()) {
		&dienicely("Couldn't logout from SMTP server", $smtp);
	}
}

if ($verbosity>2) { print "Hung up from pop3 server!\nCleaning up.\n"; }

### Look through instructions (but *NOT* pre and post) for items in @remove
### NB pre and post are meant for persistant rules, e.g. mailing lists
### instructions just for per message ones.

if (scalar @removeactions>0) {
	my (@ins,$i,$r,$changed,@newins);
	$changed=0;
	if (open(INST, '<instructions')) {
		@ins=<INST>;
		close(INST);
		for $r (@removeactions) {
			$r=quotemeta($r); # ick, but $r contains the
					  # regexp stuff like ^ so it
					  # needs quoted
			for ($i=0;$i<scalar @ins; $i++) {
				if ($ins[$i] =~ /$r/) {
					$ins[$i]="";
					$changed=1;
				}
			}
		}
		if ($changed) {
			if ($verbosity>2) { print "Writing out new instructions.\n"; }
			open(INST, '>instructions');
			if (@ins) { print INST @ins; }
			else { print INST ""; }
			close(INST);
		}
	}
}

print "Done.\n";
